/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Compute the image of \del^C restricted to kernelAB) (-> bdy.txt), */
/* and the kernel of this matrix (-> kernel.txt).                    */

package Homology;

import java.util.*;
import java.io.*;

public class Boundary
{
  public static void main(String argv[]) throws IOException {
    int deg, i, j, k, kernelDim;
    int mindeg, maxdeg;
    Vector kerElt, kernelBag, matrix[], mirror[];
    long[] mxelt, oldelt = null;
    long mult, subt, q;
//    long[] graphIndex;
          
    if (argv.length == 0) {
      mindeg = 2;
      maxdeg = 9;
      }
    else if (argv.length == 1) {
      mindeg = maxdeg = Float.valueOf(argv[0]).intValue();
      }
    else {
      mindeg = Float.valueOf(argv[0]).intValue();
      maxdeg = Float.valueOf(argv[1]).intValue();
      }
          
    for (deg=mindeg; deg<=maxdeg; deg++) {
      System.out.println("\nStarting degree "+deg+" ...");

        FileInputStream inkerfs = new FileInputStream("kernelAB" +
               String.valueOf(deg) + ".txt");             
        StreamTokenizer inkers = new StreamTokenizer(new InputStreamReader(inkerfs));
        inkers.lowerCaseMode(false);
        inkers.eolIsSignificant(false);
        inkers.parseNumbers();
        String inbdys = "bdyC" + String.valueOf(deg) + ".txt";
        FileOutputStream matrixfs = new FileOutputStream("bdy" + 
               String.valueOf(deg) + ".txt", false);             
        PrintWriter matrixs = new PrintWriter(matrixfs);
        FileOutputStream kernelfs = new FileOutputStream("kernel" + 
               String.valueOf(deg) + ".txt", false);             
        PrintWriter kernels = new PrintWriter(kernelfs);
        
        kernelBag = new Vector(20,20);
        kerElt = new Vector(100,100);    
        while (readLine(inkers, kerElt) == COLUMN) {
          kernelBag.addElement(kerElt.clone());
          kerElt.removeAllElements();    
          }
          
        inkerfs.close();
          
//        graphIndex = new long[kernelBag.size()];
//        for (i=0; i<kernelBag.size(); i++) {		// remember one of the graphs
//          kerElt = (Vector)kernelBag.elementAt(i);	// involved in the i-th elt
//          mxelt = (long[])kerElt.elementAt(0);		// of kernelAB(deg).txt
//          graphIndex[i] = mxelt[1];
//          }
        
        matrix = new Vector[kernelBag.size()];
        mirror = new Vector[kernelBag.size()];
        System.gc();

        System.out.print(kernelBag.size() + ": (matrix) ");
        
        for (i=0; i<kernelBag.size(); i++) {
          System.out.print(i + " ");
          matrix[i] = new Vector(100,100);
          kerElt = (Vector)kernelBag.elementAt(i);
          										// compute the matrix         
          for (j=0; j<kerElt.size(); j++) {
            mxelt = (long[])kerElt.elementAt(j);
            LinAlgUtil.addColumn(findImage(mxelt, inbdys), mxelt[0], matrix[i]);
            }
          }
        System.out.print("(elimination) ");

        kernelDim = 0;
        for (i=0; i<mirror.length; i++) {
          mirror[i] = new Vector(100, 100);
          mxelt = new long[2];						// the identity matrix
          mxelt[0] = 1;  mxelt[1] = i;
          mirror[i].addElement(mxelt);
          } 
          											// compute the kernel
        for (i=matrix.length-1; i>=0; i--) {
          if (matrix[i].size() == 0) {				// zero column in matrix:                                              
            										// found a kernel element
            q = LinAlgUtil.gcd(mirror[i]);
            kerElt = new Vector(100, 100);
            for (j=0; j<mirror[i].size(); j++) {
              mxelt = (long[])mirror[i].elementAt(j);
              Vector tmpVect = (Vector)kernelBag.elementAt((int)mxelt[1]);
              LinAlgUtil.addColumn(tmpVect, mxelt[0]/q, kerElt);
              }
//            kernels.print(graphIndex[i] + ": ");
            for (j=0; j<kerElt.size(); j++) {
              mxelt = (long[])kerElt.elementAt(j);
              kernels.print(mxelt[0]+" "+mxelt[1]+" "+mxelt[2]+"  ");
              }
            kernels.println(";");
            kernelDim++;
            System.out.print(i + "+ ");
            }
          else {  
            q = LinAlgUtil.gcd(matrix[i]);
//            matrixs.print(graphIndex[i] + ": ");        
            for (j=0; j<matrix[i].size(); j++) {	// print the (non-0) row
              mxelt = (long[])matrix[i].elementAt(j);
              matrixs.print(mxelt[0]/q+" "+mxelt[1]+" "+mxelt[2]+"  ");
              }
            matrixs.println(";");
            
            mxelt = (long[])matrix[i].elementAt(0); 
            for (j=i-1; j>=0; j--) {				// check each previous
              boolean occurs = false;				// column if the main 
              for (k=0; k<matrix[j].size(); k++) {	// coeff of matrix[j]
                oldelt = (long[])(matrix[j].elementAt(k));	// occurs in it
                if (mxelt[1] == oldelt[1] && mxelt[2] == oldelt[2]) {
                  occurs = true; 
                  break;
                  }
                }
              if (occurs) {
                mult = mxelt[0];
                subt = oldelt[0];
                q = LinAlgUtil.gcd(mult, subt);
                mult = mult/q;
                subt = subt/q;
                LinAlgUtil.multiply(matrix[j], mult);
                LinAlgUtil.addColumn(matrix[i], -subt, matrix[j]);
                LinAlgUtil.multiply(mirror[j], mult);
                LinAlgUtil.addColumn(mirror[i], -subt, mirror[j]);
                LinAlgUtil.coReduce(matrix[j], mirror[j]);
                }
              }
            System.out.print(i + " ");
            }
          }        
        
        kernels.flush();  kernels.close();   kernelfs.close();
        matrixs.flush();  matrixs.close();   matrixfs.close();
        System.out.println("kernel = " + kernelDim + " image = " +
            (matrix.length - kernelDim) + "; done "+new Date());
        }
    }
        
  final static public int COLUMN = -2;
  final static public int EOF = -1;	// non-neg values mean "Graph X" lines
  
  public static int readLine(StreamTokenizer s, Vector bag)
                                                      throws IOException {
    long mxelt[] = new long[3];
       
    s.nextToken();
    if (s.ttype == StreamTokenizer.TT_EOF)
      return EOF;
      
    while (s.ttype == StreamTokenizer.TT_NUMBER) {
      mxelt[0] = (long)s.nval;
      s.nextToken();
      mxelt[1] = (long)s.nval;
      s.nextToken();
      mxelt[2] = (long)s.nval;
      s.nextToken();
      bag.addElement(mxelt.clone());
      }
    
    if (s.ttype == (int)';')
      return COLUMN;
    
    if (s.ttype == StreamTokenizer.TT_WORD && s.sval.equals("Graph")) {
      s.nextToken();
      return (int)s.nval;
      }
    
    System.out.println("Error reading "+s);
    System.exit(1);
    return EOF;		// well, it doesn't, but the compiler requires it
    }      
  
    
  public static Vector findImage(long[] mxelt, String filename) 
                                                   throws IOException {
    int i;
    Vector bag = new Vector(100,100);
    
      FileInputStream fs = new FileInputStream(filename);             
      StreamTokenizer s = new StreamTokenizer(new InputStreamReader(fs));
      s.lowerCaseMode(false);
      s.eolIsSignificant(false);
      s.parseNumbers();
	 
	  do {					            		// find "Graph x" string
	    s.nextToken();
	    } while (s.ttype != StreamTokenizer.TT_WORD ||
	             ! s.sval.equals("Graph") ||
	             s.nextToken() != StreamTokenizer.TT_NUMBER ||
	             (long)s.nval != mxelt[1]);
	                
	  for (i=0; i<mxelt[2]; i++) {				// find yth line (separated
	    do {									//           by semicolons)
	      s.nextToken();
	      } while (s.ttype != (int)';');
	    }
	      
	  readLine(s, bag);

    fs.close();	        
    return bag;
    }
            
}    
