/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Compute \del^A of the graphs.                                   */

package Homology;

import java.util.*;
import java.io.*;

public class BoundaryA
{
  public static void main(String argv[]) throws IOException,
                                                ClassNotFoundException {
    int deg, i, j, k, p, num, kernelDim,
        mindeg, maxdeg;
    long mult, subt, q;
    Graph g;
    Tree t, t1;
    Vector matrix[], treeBag, mirror[];
    long[] mxelt = new long[2], oldelt = new long[2];
    boolean found;
          
    if (argv.length == 0) {
      mindeg = 2;
      maxdeg = 9;
      }
    else if (argv.length == 1) {
      mindeg = maxdeg = Float.valueOf(argv[0]).intValue();
      }
    else {
      mindeg = Float.valueOf(argv[0]).intValue();
      maxdeg = Float.valueOf(argv[1]).intValue();
      }
          
    for (deg=mindeg; deg<=maxdeg; deg++) {
      kernelDim = 0;
      System.out.print("Starting degree "+deg+" ... ");
      FileInputStream infs = new FileInputStream("nbag" +
             String.valueOf(deg) + ".out");             
      ObjectInputStream ins = new ObjectInputStream(infs);
      FileOutputStream matrixfs = new FileOutputStream("bdyA" + 
             String.valueOf(deg) + ".txt", false);             
      PrintWriter matrixs = new PrintWriter(matrixfs);
      FileOutputStream kernelfs = new FileOutputStream("kernelA" + 
             String.valueOf(deg) + ".txt", false);             
      PrintWriter kernels = new PrintWriter(kernelfs);

      i = 0;
      while (true) {
        try { 
          g = (Graph)ins.readObject();     
          treeBag = new Vector(100, 100);
          matrix = new Vector[g.treeBag.size()];
          System.gc();
          System.out.print(i);
          matrixs.println("Graph "+i);

          for (j=0; j<matrix.length; j++) {		// Construct the matrix:
            matrix[j] = new Vector(100, 100);
            t = new Tree((byte[])g.treeBag.elementAt(j));
            t.parent = g;
            for (k=0; k<t.noEdges; k++) {
              t1 = t.removeEdge(k);
              if (t1.hasOddSymmetry())
                continue;
              mxelt = addToBag(treeBag, t1);	// find index of t1
              if (k%2 == 0)
                mxelt[0] = -mxelt[0];			// multiply coeff by (-1)^m
              found = false;
              for (p=0; p<matrix[j].size(); p++) {
                oldelt = (long[])matrix[j].elementAt(p);
                if (mxelt[1] == oldelt[1]) {
                  if ((oldelt[0] += mxelt[0]) == 0)	// increase coefficient
                    matrix[j].removeElementAt(p);	// if tree occurred in
                  found = true;						// this column already...
                  break;
                  }
                }
              if (!found)							// ... add new element
                matrix[j].addElement(mxelt.clone());// to column if not.
              }
            } 
          
          for (j=0; j<matrix.length; j++) {
            for (k=0; k<matrix[j].size(); k++) {
              mxelt = (long[])matrix[j].elementAt(k);
              matrixs.print(mxelt[0]+" "+mxelt[1]+"  ");
              }
            matrixs.println(";");
            }

          mirror = new Vector[g.treeBag.size()];
          for (j=0; j<mirror.length; j++) {
            mirror[j] = new Vector(100, 100);
            mxelt = new long[2];					// the identity matrix
            mxelt[0] = 1;  mxelt[1] = j;
            mirror[j].addElement(mxelt);
            } 

//System.out.println(i+" "+matrix.length);

          num = 0;         
          for (j=0; j<matrix.length; j++) {        	// compute the kernel

            if (matrix[j].size() == 0) {			// zero column in matrix:
              if (mirror[j].size() > 0) {			// found a kernel element
                LinAlgUtil.reduce(mirror[j]);
                for (k=0; k<mirror[j].size(); k++) {
                  mxelt = (long[])mirror[j].elementAt(k);
                  kernels.print(mxelt[0]+" "+ i +" "+mxelt[1]+"  ");
                  }
                }
              kernels.println(";");
              num++;
              continue;
              }
              
            mxelt = (long[])matrix[j].elementAt(0); 
            for (k=j+1; k<matrix.length; k++) {		// check each subsequent
              boolean occurs = false;				// column if the main 
              for (p=0; p<matrix[k].size(); p++) {	// coeff of matrix[j]
                oldelt = (long[])(matrix[k].elementAt(p));	// occurs in it
                if (mxelt[1] == oldelt[1]) {
                  occurs = true; 
                  break;
                  }
                }
              if (occurs) {
                mult = mxelt[0];
                subt = oldelt[0];
                q = LinAlgUtil.gcd(mult, subt);
                mult = mult/q;
                subt = subt/q;
                LinAlgUtil.multiply(matrix[k], mult);
                LinAlgUtil.addColumn(matrix[j], -subt, matrix[k]);
                LinAlgUtil.multiply(mirror[k], mult);
                LinAlgUtil.addColumn(mirror[j], -subt, mirror[k]);
                LinAlgUtil.coReduce(matrix[k], mirror[k]);
                }
              }
            }

          System.out.print("("+num+") ");
          kernelDim += num;         
          i++;
          }
        catch (EOFException e) {
          break;
          }
        }
        
      kernels.flush();  kernels.close();   kernelfs.close();
      matrixs.flush();  matrixs.close();   matrixfs.close();
      ins.close();  infs.close();
      System.out.println("dim = "+kernelDim+"; done "+new Date());
      }
    }


  private static long[] addToBag(Vector bag, Tree t1) {
    int i, j;
    Tree t2;
    Permutation sigma;
    long[] coeff = new long[2];
   
    for (i=0; i<bag.size(); i++) {
      t2 = (Tree)bag.elementAt(i);
      if ((sigma = t1.findIsomorphism(t2)) != null) {
          coeff[0] = (long) t1.sign(sigma);
          coeff[1] = (long) i;					// the sign of the permutation
          return coeff;					// which takes t1.permuted(sigma) 
          }								// into t2 (i.e., normal form)
       }
      
    coeff[0] = 1;			// not in the bag yet
    coeff[1] = (long) bag.size();
    bag.addElement(new Tree(t1));
    return coeff;
    }
}

