/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Compute the image of \del^B restricted to ker(\del^A) (-> bdyAB.txt), */
/* and the kernel of this matrix (-> kernelAB.txt).                      */

package Homology;

import java.util.*;
import java.io.*;

public class BoundaryAB
{
  public static void main(String argv[]) throws IOException {
    int deg, i, j, k, kernelDim;
    long mult, subt, q;
    int mindeg, maxdeg;
    String bdyB;
    Vector kerElt, kernelBag, matrix[], mirror[];
    long[] mxelt, oldelt = null;
    FileOutputStream kernelfs;
    PrintWriter kernels;
          
    if (argv.length == 0) {
      mindeg = 2;
      maxdeg = 9;
      }
    else if (argv.length == 1) {
      mindeg = maxdeg = Float.valueOf(argv[0]).intValue();
      }
    else {
      mindeg = Float.valueOf(argv[0]).intValue();
      maxdeg = Float.valueOf(argv[1]).intValue();
      }
          
    for (deg=mindeg; deg<=maxdeg; deg++){
      System.out.println("\nStarting degree "+deg+" ...");
        FileInputStream inkerfs = new FileInputStream("kernelA" +
               String.valueOf(deg) + ".txt");             
        StreamTokenizer inkers = new StreamTokenizer(new InputStreamReader(inkerfs));
        inkers.lowerCaseMode(false);
        inkers.eolIsSignificant(false);
        inkers.parseNumbers();
        String inbdys = "bdyB" + String.valueOf(deg) + ".txt";
        
        kernelBag = new Vector(20,100);
        kerElt = new Vector(100,100);    
        while (readLine(inkers, kerElt, 3) == Boundary.COLUMN) {
          kernelBag.addElement(kerElt.clone());
          kerElt.removeAllElements();    
          }
        inkerfs.close();
        matrix = new Vector[kernelBag.size()];
        mirror = new Vector[kernelBag.size()];
        System.gc();

        System.out.print(kernelBag.size() + ": ");
        
        for (i=0; i<kernelBag.size(); i++) {
          System.out.print(i + " ");
          matrix[i] = new Vector(100, 100);
          kerElt = (Vector)kernelBag.elementAt(i);
          										// compute the matrix         
          for (j=0; j<kerElt.size(); j++) {
            mxelt = (long[])kerElt.elementAt(j);
            LinAlgUtil.addColumn(findImage(mxelt, inbdys, 2), mxelt[0], matrix[i]);
            }
          }
          
        kernelDim = 0;
        for (i=0; i<mirror.length; i++) {
          mirror[i] = new Vector(100, 100);
          mxelt = new long[2];					// the identity matrix
          mxelt[0] = 1;  mxelt[1] = i;
          mirror[i].addElement(mxelt);
          } 
          
        for (i=0; i<matrix.length; i++) {      	// compute the kernel
          if (matrix[i].size() == 0) {          
            kerElt = new Vector(100, 100);			// zero column in matrix:  
            for (j=0; j<mirror[i].size(); j++) {	// we've found a kernel element
              mxelt = (long[])mirror[i].elementAt(j);
              Vector tmpVect = (Vector)kernelBag.elementAt((int)mxelt[1]);	// ???
              LinAlgUtil.addColumn(tmpVect, mxelt[0], kerElt);
              }
            LinAlgUtil.reduce(kerElt);
            
            kernelfs = new FileOutputStream("kernelAB" + 
                                          String.valueOf(deg) + ".txt", false);             
            kernels = new PrintWriter(kernelfs);
            for (j=0; j<kerElt.size(); j++) {
              mxelt = (long[])kerElt.elementAt(j);
              kernels.print(mxelt[0]+" "+mxelt[1]+" "+mxelt[2]+"  ");
              }
            kernels.println(";");
            kernels.flush();  kernels.close();   kernelfs.close();
            kernelfs = null;   kernels = null;            
            
            kernelDim++;
            matrix[i] = null;   mirror[i] = null;
            System.gc();
            continue;
            }
            
          mxelt = (long[])matrix[i].elementAt(0); 
          for (j=i+1; j<matrix.length; j++) {		// check each subsequent
            boolean occurs = false;					// column if the main 
            for (k=0; k<matrix[j].size(); k++) {	// coeff of matrix[j]
              oldelt = (long[])(matrix[j].elementAt(k));	// occurs in it
              if (mxelt[1] == oldelt[1]) {
                occurs = true; 
                break;
                }
              }
            if (occurs) {
              mult = mxelt[0];
              subt = oldelt[0];
              q = LinAlgUtil.gcd(mult, subt);
              mult = mult/q;
              subt = subt/q;
              LinAlgUtil.multiply(matrix[j], mult);
              LinAlgUtil.addColumn(matrix[i], -subt, matrix[j]);
              LinAlgUtil.multiply(mirror[j], mult);
              LinAlgUtil.addColumn(mirror[i], -subt, mirror[j]);
              LinAlgUtil.coReduce(matrix[j], mirror[j]);
              }
            }
          matrix[i] = mirror[i] = null;
          System.gc();
          }      
        
        System.out.println("dim = "+kernelDim+"; done "+new Date());
      }
    }
        
//  final static private int COLUMN = -2;
//  final static private int EOF = -1;	// non-neg values mean "Graph X" lines
  
  public static long readLine(StreamTokenizer s, Vector bag, int coords)
                                                      throws IOException {
    long mxelt[] = new long[coords], n;
       
    s.nextToken();
    if (s.ttype == StreamTokenizer.TT_EOF)
      return Boundary.EOF;
      
    while (s.ttype == StreamTokenizer.TT_NUMBER) {
      mxelt[0] = (long)s.nval;
      s.nextToken();
      mxelt[1] = (long)s.nval;
      s.nextToken();
      if (coords == 3) {		// coords == 2 for bdyB, 3 for kernelA
        mxelt[2] = (long)s.nval;
        s.nextToken();
        }
      bag.addElement(mxelt.clone());
      }
     
    if (s.ttype == (int)';')
      return Boundary.COLUMN;
    
    if (s.ttype == StreamTokenizer.TT_WORD && s.sval.equals("Graph")) {
      s.nextToken();
      return (int)s.nval;
      }
    
    System.out.println("Error reading "+s);
    System.exit(1);
    return Boundary.EOF;
    }      
  
    
  public static Vector findImage(long[] mxelt, String filename, int coords) 
                                               throws IOException    {
    int i, bagElt[];
    Vector bag = new Vector(100,100);
    
    FileInputStream fs = new FileInputStream(filename);             
    StreamTokenizer s = new StreamTokenizer(new InputStreamReader(fs));
	 
	  do {					            		// find "Graph x" string
	    s.nextToken();
	    } while (s.ttype != StreamTokenizer.TT_WORD ||
	             ! s.sval.equals("Graph") ||
	             s.nextToken() != StreamTokenizer.TT_NUMBER ||
	             (long)s.nval != mxelt[1]);
	                
	  for (i=0; i<mxelt[2]; i++) {				// find yth line (separated
	    do {									//           by semicolons)
	      s.nextToken();
	      } while (s.ttype != (int)';');
	    }
	      
	  readLine(s, bag, coords);		// coords = 2 always here... 

    fs.close();	        
    return bag;
    }
            
}    
