/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Compute \del^C of the graphs.                                   */

package Homology;

import java.util.*;
import java.io.*;

public class BoundaryC
{
  public static void main(String argv[]) throws IOException,
                                                ClassNotFoundException {
    int deg, i, j, k, p, num,
        mindeg, maxdeg;
    Graph g, graphs[], graphN[];
    Permutation permuts[];
    Tree t, t1;
    Vector matrix[], graphBag;
    long[] mxelt = new long[3], oldelt = new long[3], subelt = new long[3];
    boolean found;
    
    if (argv.length == 0) {
      mindeg = 3;
      maxdeg = 9;
      }
    else if (argv.length == 1) {
      mindeg = maxdeg = Float.valueOf(argv[0]).intValue();
      }
    else {
      mindeg = Float.valueOf(argv[0]).intValue();
      maxdeg = Float.valueOf(argv[1]).intValue();
      }
          
    for (deg=mindeg; deg<=maxdeg; deg++) {
      System.out.print("Starting degree "+deg+" ... ");
      num = 0;

      FileInputStream newinfs = new FileInputStream("nbag" +
             String.valueOf(deg) + ".out");             
      ObjectInputStream newins = new ObjectInputStream(newinfs);
      FileInputStream oldinfs = new FileInputStream("nbag" +
             String.valueOf(deg-1) + ".out");             
      ObjectInputStream oldins = new ObjectInputStream(oldinfs);
      FileOutputStream outfs = new FileOutputStream("bdyC" + 
             String.valueOf(deg) + ".txt", false);             
      PrintWriter outs = new PrintWriter(outfs);
        
      graphBag = new Vector(100, 100);
      while (true) {
        try {
          graphBag.addElement(oldins.readObject());
          }
        catch (EOFException e) {
          break;
          }
        } 
         
      i = 0;   
      while (true) {
        try {
          g = (Graph)newins.readObject();     
          matrix = new Vector[g.treeBag.size()];
          graphs = new Graph[g.noEdges];	// contracted graphs a la nature
          graphN = new Graph[g.noEdges];	// contr. graphs in normal form
          permuts = new Permutation[g.noEdges];
          System.gc();
          System.out.print(i+" ");
          outs.println("Graph "+i);

          t = new Tree(g);
          for (j=0; j<g.noEdges; j++) {			// table of all possible
            t1 = t.contractEdge(j, true);		// contracted graphs for
            if (t1 != null) {					// this parent graph g
              graphs[j] = t1.parent;
              for (k=1; k<t1.size; k++) {
                t1.m[0][k] = (byte)k;
                for (p=k+1; p<t1.size; p++)
                  t1.m[k][p] = 0;
                }
              t1.noEdges = t1.size - 1;
              t1 = t1.normalForm(false);	
              graphN[j] = t1.parent;
              permuts[j] = new Permutation(t1.size);// record the permutation
              for (k=1; k<t1.size; k++)			// which takes the contracted 
                permuts[j].p[k] = t1.m[0][k];	// graph to normal form
              }
            }

          for (j=0; j<matrix.length; j++) {		// Construct the matrix:
            matrix[j] = new Vector(100, 100);
            t = new Tree((byte[])g.treeBag.elementAt(j));
            t.parent = g;
            for (k=0; k<t.noEdges; k++) {
              if ((t1 = t.contractEdge(k, true)) == null)	// contract edge
                continue;					// if it is at the basepoint
              for (p=0; p<g.noEdges; p++)
                if (graphs[p]!=null && t1.parent.isIsomorphic(graphs[p]))
                  break;
              t1.parent = graphN[p];
              t1 = t1.permuted(permuts[p]);
              mxelt = findInBag(graphBag, t1);	// find index of t1
              if (mxelt == null)	// not in graphBag: could be because of
                continue;			// odd symmetry, Lemma 4.3 or 4.4
              if (k%2 != 0)
                mxelt[0] = -mxelt[0];			// multiply coeff by (-1)^m
              mxelt[0] *= t1.coeff;				// sundry (-1)'s
              found = false;
              for (p=0; p<matrix[j].size(); p++) {
                oldelt = (long[])matrix[j].elementAt(p);
                if (mxelt[1] == oldelt[1] && mxelt[2] == oldelt[2]) {
                  if ((oldelt[0] += mxelt[0]) == 0)	// increase coefficient
                    matrix[j].removeElementAt(p);	// if tree occurred in
                  found = true;						// this column already...
                  break;
                  }
                }
              if (!found)							// ... add new element
                matrix[j].addElement(mxelt.clone());// to column if not.
              }
            }
          
          for (j=0; j<matrix.length; j++) {
            for (k=0; k<matrix[j].size(); k++) {
              mxelt = (long[])matrix[j].elementAt(k);
              outs.print(mxelt[0]+" "+mxelt[1]+" "+mxelt[2]+"  ");
              }
            outs.println(";");
            }
          i++;
          }
        catch (EOFException e) {
          break;
          }
        }

      oldins.close();   oldinfs.close();
      newins.close();   newinfs.close();
      outs.flush();  outs.close();  outfs.close();

      System.out.println("done "+new Date());
      }
    }


  private static long[] findInBag(Vector bag, Tree t1) {
    int i, j;
    Graph g;
    Tree t2;
    Permutation sigma;
    long[] coeff = new long[3];
   
    for (i=0; i<bag.size(); i++) {
      g = (Graph)bag.elementAt(i);
      for (j=0; j<g.treeBag.size(); j++) {
        t2 = new Tree((byte[])g.treeBag.elementAt(j));
        t2.parent = g;
        if (t1.parent.equals(t2.parent)) {
          if ((sigma = t2.findIsomorphism(t1)) != null) {
            coeff[0] = (long) t2.sign(sigma);
            coeff[1] = (long) i;
            coeff[2] = (long) j;		  	// the sign of the permutation
            return coeff;					// which takes t2.permuted(sigma) 
            }								// into t1
          }
        }
      }
      
    return null;	// return null if not in the bag; that means either
    }				// it was collapsed earlier
    
}

