/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* LinAlgUtil class                                                */
/* Utility package for dealing with matrices                       */
/* (identical to OutH.LinAlgUtil used in the Out computation)      */

package Homology;

import java.util.Vector;

public class LinAlgUtil
{

  public static void multiply(Vector bag, long number) {
    int i;
    long mxelt[];
    
    for (i=0; i<bag.size(); i++) {		// multiply bag-column by number
      mxelt = (long[])bag.elementAt(i);
      mxelt[0] *= number;
      }
    }
    
    
  public static void divide(Vector bag, long number) {
    int i;
    long mxelt[];
    
    for (i=0; i<bag.size(); i++) {		// divide bag-column by number
      mxelt = (long[])bag.elementAt(i);
      mxelt[0] /= number;
      }
    }
    
    
  public static void addColumn(Vector what, long number, Vector whto) {
    int i, j,  coords;
    long mxelt[], oldelt[];
    boolean found;
    
    if (what.size() == 0)
      return;
    else {
      mxelt = (long[])what.elementAt(0);
      coords = mxelt.length;	// 2 or 3, depending on the situation
      }
    
    for (i=0; i<what.size(); i++) {			// subtract number * 'what'-
      mxelt = (long[])what.elementAt(i);	// column from 'from'-column
      found = false;
      for (j=0; j<whto.size(); j++) {
        oldelt = (long[])whto.elementAt(j);
        if (mxelt[1]==oldelt[1] && (coords==2 || mxelt[2]==oldelt[2])) {
          found = true;
          if ((oldelt[0] += number * mxelt[0]) == 0)
            whto.removeElementAt(j);
          break;
          }
        }
      if (!found) {
        oldelt = (long[])mxelt.clone();
        oldelt[0] *= number;
        whto.addElement(oldelt);
        }
      }
    }


  public static void reduce(Vector column) {	// divide the vector by the
    long q;										//   gcd of its elements
    
    q = gcd(column);
    if (q > 1)
      divide(column, q);
    }    


  public static void coReduce(Vector one, Vector two) {		// divide both
    long q;													// colums by their
    														// common factor
    q = gcd(one);
    q = gcd(q, gcd(two));
    if (q > 1) {
      divide(one, q);
      divide(two, q);
      }
    }
    
   
  public static long gcd(Vector column) {		// find the gcd of the 
    int i;										//  elements of a Vector
    long mxelt[], q;
    
    if (column.size() == 0)
      return 1;
    
    q = 0;
    for (i=0; i<column.size(); i++) {
      mxelt = (long[])column.elementAt(i);
      q = gcd(mxelt[0], q);
      }
      
    return q;
    }    


  public static long gcd(long a, long b) {
    long r;							// for some reason, this basic
    								// gcd function is not in the
   	if (a < 0)  a = -a;				// standard java libraries
   	if (b < 0)  b = -b;				// (at least I couldn't find it)
    if (a == 0)  return b;
    if (b == 0)  return a;
    if (a < b) { r = a; a = b; b = r; }
    
    while (b != 0) {
      r = a%b;
      a = b;
      b = r;
      }
    return a;
    }

}
