/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* ListGraphs class                                                */
/* List all automorphism classes of graphs up to degree 9          */
/*   (but no subtrees or automorphisms yet)                        */
/*   note: does NOT work for degrees higher than 9                 */

package Homology;

import java.util.Vector;
import java.io.*;

public class ListGraphs 
{
  private final static int[] count = { 0, 0, 1, 2, 5, 9, 26, 59, 197, 547 };
											// number of graphs in each degree	
  public static void main(String argv[]) {
    int i, j, k, deg, n,
        mindeg, maxdeg;
    boolean isNew, hasNoSeparatingEdges;
    Graph oldGraph, newGraph;
    Tree tmpTree, tmpTree2;
    Vector oldBag, newBag = new Vector(10,10);

    if (argv.length == 0) {
      mindeg = 2;
      maxdeg = 9;
      }
    else if (argv.length == 1) {
      mindeg = maxdeg = Float.valueOf(argv[0]).intValue();
      }
    else {
      mindeg = Float.valueOf(argv[0]).intValue();
      maxdeg = Float.valueOf(argv[1]).intValue();
      }

    if (mindeg == 2) {
      Graph basicGraph = new Graph(2);	// one degree 1 graph only
      basicGraph.m[0][1] = 3;
      basicGraph.neighbors = 1;
      basicGraph.valency = 3;
      basicGraph.noEdges = 1;
      newBag.addElement(basicGraph);
      }
    else { try {
      FileInputStream fs = new FileInputStream("bag" +
              String.valueOf(mindeg-1) + ".out");
      ObjectInputStream s = new ObjectInputStream(fs);
      for (i=0; i<count[mindeg-1]; i++) {
        oldGraph = (Graph)s.readObject();
        oldGraph.treeBag = null;
        oldGraph.symBag = null;
        newBag.addElement(oldGraph);
        }
      s.close();
      fs.close();
      }
    catch (IOException e) {
      System.out.println("Error reading file \"bag"+
                          String.valueOf(mindeg-1)+".out\".");
      System.exit(1);
      }
    catch (ClassNotFoundException e) {
      System.out.println("Class not found.");
      System.exit(1);
      }
    }     
          
    for (deg=mindeg; deg<=maxdeg; deg++) {

      oldBag = newBag;
      newBag = new Vector(10,10);

      for (n=0; n<oldBag.size(); n++) {

        oldGraph = (Graph)oldBag.elementAt(n);
        System.out.print(n+"("+oldGraph.neighbors+")");
       
        for (i=0; i<oldGraph.neighbors; i++) {		// new graphs generated by the 
          newGraph = oldGraph.plusedge(i);			//   "plus edge" operation
          newGraph.toNormalForm();
          isNew = true;
          for (k=0; k<newBag.size(); k++)
            if (newGraph.isIsomorphic((Graph)newBag.elementAt(k))) {
              isNew = false;
              //System.out.print(".");
              break;
              }
          if (isNew) {
            newGraph.treeBag = null;
            newGraph.symBag = null;
            newBag.addElement(newGraph);
            //System.out.print("+");
            }
          }
        
        for (i=0; i<oldGraph.neighbors; i++)		// new graphs generated by the
          for (j=i+1; j<oldGraph.neighbors; j++) {  //   "glue edges" operation
            newGraph = oldGraph.glueedge(i,j);
            
            hasNoSeparatingEdges = true;
            tmpTree = new Tree(newGraph);
            for (k=0; k<tmpTree.noEdges; k++) {		// check whether the new graph
              tmpTree2 = tmpTree.removeEdge(k);		// has a separating edge
              if (!tmpTree2.isConnected()) {
                hasNoSeparatingEdges = false;
                break;
                }
              }
            if (hasNoSeparatingEdges) {
              newGraph.toNormalForm();            
              isNew = true;
              for (k=0; k<newBag.size(); k++)
                if (newGraph.isIsomorphic((Graph)newBag.elementAt(k))) {
                  isNew = false;
                  //System.out.print(".");
                  break;
                  }
              if (isNew)	{
                newGraph.treeBag = null;
                newGraph.symBag = null;
                newBag.addElement(newGraph);
                //System.out.print("+");
                }
              }
            }
          }  // end of 'n' loop
        
      if (deg == 8) {			// add reducible graphs in degree 8
        Graph[] fourBag = new Graph[count[4]];
        try {
          FileInputStream in4fs = new FileInputStream("bag4.out");
          ObjectInputStream in4s = new ObjectInputStream(in4fs);
          for (i=0; i<count[4]; i++) {
            oldGraph = (Graph)in4s.readObject();	// read degree 4 graphs
            oldGraph.treeBag = null;
            oldGraph.symBag = null;
            fourBag[i] = oldGraph;
            }
          in4s.close();
          in4fs.close();
          }
        catch (IOException e) {
          System.out.println("Error reading file \"bag4.out\".");
          System.exit(1);
         }
        catch (ClassNotFoundException e) {
          System.out.println("Class not found.");
          System.exit(1);
          }
        
        for (i=0; i<count[4]; i++)
          for (j=i; j<count[4]; j++) {
            newGraph = amalgamate(fourBag[i], fourBag[j]);
            newGraph.toNormalForm();				// form all possible wedges
            isNew = true;
            for (k=0; k<newBag.size(); k++)
              if (newGraph.isIsomorphic((Graph)newBag.elementAt(k))) {
                isNew = false;
                //System.out.print(".");
                break;
                }
            if (isNew)	{
              newGraph.treeBag = null;
              newGraph.symBag = null;
              newBag.addElement(newGraph);			// add them to the bag
              //System.out.print("+");
              }
            }
          }

      if (deg == 9) {			// add reducible graphs in degree 9
        Graph[] fourBag = new Graph[count[4]];
        Graph[] fiveBag = new Graph[count[5]];
        try {
          FileInputStream in4fs = new FileInputStream("bag4.out");
          ObjectInputStream in4s = new ObjectInputStream(in4fs);
          for (i=0; i<count[4]; i++) {
            oldGraph = (Graph)in4s.readObject();	// read degree 4 graphs
            oldGraph.treeBag = null;
            oldGraph.symBag = null;
            fourBag[i] = oldGraph;
            }
          in4s.close();
          in4fs.close();
          FileInputStream in5fs = new FileInputStream("bag5.out");
          ObjectInputStream in5s = new ObjectInputStream(in5fs);
          for (i=0; i<count[5]; i++) {
            oldGraph = (Graph)in5s.readObject();	// read degree 5 graphs
            oldGraph.treeBag = null;
            oldGraph.symBag = null;
            fiveBag[i] = oldGraph;
            }
          in5s.close();
          in5fs.close();
          }
        catch (IOException e) {
          System.out.println("Error reading file \"bag4.out\" or \"bag5.out\".");
          System.exit(1);
         }
        catch (ClassNotFoundException e) {
          System.out.println("Class not found.");
          System.exit(1);
          }
        
        for (i=0; i<count[4]; i++)
          for (j=0; j<count[5]; j++) {
            newGraph = amalgamate(fourBag[i], fiveBag[j]);
            newGraph.toNormalForm();				// form all possible wedges
            isNew = true;
            for (k=0; k<newBag.size(); k++)
              if (newGraph.isIsomorphic((Graph)newBag.elementAt(k))) {
                isNew = false;
                //System.out.print(".");
                break;
                }
            if (isNew)	{
              newGraph.treeBag = null;
              newGraph.symBag = null;
              newBag.addElement(newGraph);			// add them to the bag
              //System.out.print("+");
              }
            }
          }
      
      try{
        FileOutputStream outfs = new FileOutputStream("bag" +
                                       String.valueOf(deg) + ".out", false);
        ObjectOutputStream outs = new ObjectOutputStream(outfs);
        for (i=0; i<newBag.size(); i++) {
          newGraph = (Graph)newBag.elementAt(i);
          outs.writeObject(newGraph);
          }
        outs.flush();
        outs.close();
        outfs.close();
        }
      catch (IOException e) {
        System.out.println("Error writing graph #"+newBag.size()
                                    +" of degree "+deg+".");
        System.exit(1);
        }
      System.out.println("\n"+newBag.size()+" graphs of degree "+deg +
                                        " written: "+new java.util.Date());
      System.out.println();
      }
    }
    
    
  private static Graph amalgamate(Graph g1, Graph g2) {
    int i, j, k, offset;							// form the wedge product
    Graph g = new Graph(g1.size + g2.size - 1);		//  of graphs g1 and g2
    
    for (i=0; i<g1.size; i++)
      for (j=i+1; j<g1.size; j++)					// copy g1
        g.m[i][j] = g1.m[i][j];
    
    offset = g1.size - 1;
    for (j=1; j<g2.size; j++)						// write the neighbors of
      g.m[0][offset + j] = g2.m[0][j];				// the basepoint in g2
    
    for (i=1; i<g2.size; i++)
      for (j=i+1; j<g2.size; j++)					// all other edges of g2
        g.m[offset + i][offset + j] = g2.m[i][j];
    
    g.valency = g1.valency + g2.valency;
    g.neighbors = g1.neighbors + g2.neighbors;
    g.noEdges = g1.noEdges + g2.noEdges;
    
    return g;
    }
    
}

