/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* ListTrees class                                                 */
/* Finds the subtrees of the graphs.                               */

package Homology;

import java.util.*;
import java.io.*;

public class ListTrees
{
  public static void main(String argv[]) throws IOException, 
                                                ClassNotFoundException {
    int i, j, k,
        number, min, max, sum,
        deg, mindeg, maxdeg;
    Graph graph;
    Tree t, t1;
    int[] subset;
    Vector treeBag;

    if (argv.length == 0) {
      mindeg = 2;
      maxdeg = 9;
      }
    else if (argv.length == 1) {
      mindeg = maxdeg = Float.valueOf(argv[0]).intValue();
      }
    else {
      mindeg = Float.valueOf(argv[0]).intValue();
      maxdeg = Float.valueOf(argv[1]).intValue();
      }
          
    for (deg=mindeg; deg<=maxdeg; deg++) {
      System.out.print("\nStarting degree "+deg+" ... ");
      treeBag = new Vector(100,100);
      min = 10000;
      max = 0;
      sum = 0;
      FileInputStream infs = new FileInputStream("bag" +
             String.valueOf(deg) + ".out");
      ObjectInputStream ins = new ObjectInputStream(infs);
      FileOutputStream outfs = new FileOutputStream("nbag" +
             String.valueOf(deg) + ".out", false);
      ObjectOutputStream outs = new ObjectOutputStream(outfs);

      number = j = 0;
      while (true) {
        try {
          graph = (Graph)ins.readObject();
          graph.treeBag = new Vector(Graph.bagSize[deg],Graph.bagSize[deg]);
          graph.symBag = new Vector(5,5);
          graph.toNormalForm();				// toNormalForm() restores the list of

          number++;							// symmetries we threw away earlier to
          subset = new int[graph.noEdges];	// conserve memory in ListGraphs
          
          for (k=0; k<deg; k++)		// initialize subset to first
            subset[k] = 1;			// 'deg' (=#vertices-1) edges
          
          do {
            t = graph.subGraph(subset);
            if (t.isConnected() && !t.hasOddSymmetry()	
            						// orientation-reversing automorphism
                                && !t.containsBadTriangle())
                                	// Hatcher-Vogtmann's Lemma 4.3 or 4.4 
              addTree(treeBag, t);
              						// addTree checks whether the tree is new
            } while (nextSubset(subset));

          for (k=0; k<treeBag.size(); k++) {
            t = (Tree)treeBag.elementAt(k);
            graph.treeBag.addElement(t.prfer());
            }
          
          outs.writeObject(graph);
          graph = null;
          treeBag.removeAllElements();
          System.gc();

          if (k<min)
            min = k;
          if (k>max)
            max = k;
          sum += k;
          System.out.print("("+j+","+k+") ");
          j++;
          }
        catch (EOFException e) {
          break;
          }
        }
          
        ins.close();   infs.close();
        outs.flush();  outs.close();  outfs.close();
        System.out.println("done. "+new Date()+"\nmin: "+min+
            ", max: "+max+", sum: "+sum+" (out of "+number+")");
        }
      }
    
    
  private static boolean nextSubset(int[] subset) {
    int i, j, k;
    boolean isLast = true;
    
    k = -1;
    for (i=0; i<subset.length-1; i++) {
      if (subset[i] == 1) {				// find first occurrence
        k++;							// of a 1 followed by a 0
        if (subset[i+1] == 0) {
          isLast = false;
          break;
          }
        }
      }
    if (isLast)
      return false;
    
    for (j=0; j<k; j++)		// advance to next subset (of same
      subset[j] = 1;		// number of elements) in reverse-
    for (; j<=i; j++)		// lexicographic order
      subset[j] = 0;
    subset[j] = 1;

    return true;
    }
    
    
  private static void addTree(Vector bag, Tree tree) {
    int i;
    Tree t1;
    
    for (i=0; i<bag.size(); i++) {		// check if the tree is already
      t1 = (Tree)bag.elementAt(i);		//   in the bag
      if (tree.isIsomorphic(t1))
        return;
      }
        
    bag.addElement(tree);
    }
}
