/*                                                                 */
/* Aut(F_n) computation                                            */
/*                                                                 */
/* Ferenc Gerlits, PhD dissertation, Cornell University, Aug 2002  */
/*                                                                 */
/* Permutation class                                               */
/* with a 'next' function which steps through all permutations     */

package Homology;

public class Permutation extends Object 
                         implements java.io.Serializable
{
  public int size;
  public int p[];
  
  public Permutation(int s) {				// the identity permutation
    int i;
    
    size = s;
    p = new int[size];
    for (i=0; i<size; i++)
      p[i] = i;
    }
    
  public Permutation(Permutation sigma) {	// make a clone of sigma
    int i;
    
    size = sigma.size;
    p = new int[size];
    for (i=0; i<size; i++)
      p[i] = sigma.p[i];
    }

    
  public Permutation(int s, int[] b) {		// permutation out of an array
    int i;
    
    size = s;
    p = new int[size];
    for (i=0; i<size; i++)
      p[i] = b[i];
    }
    
    
  public int sign() {				// returns 1 for even permutations,
    int i, j, inversions = 0;		//    and -1 for odd ones
    
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        if (p[j] < p[i])
          inversions++;
    if (inversions % 2 == 0)
      return 1;
    else return -1;
    }
    
    
  public boolean next() {			// cycle through the permutations;
    int i, j, first;				// 12...n is first, n...21 is last
    int q[];
    Permutation sigma;
    
    if (size == 2) {
      if (p[0] == 0) {
        p[0] = 1;
        p[1] = 0;
        return true;
        }
      else return false;
      }
    
    first = p[0];					// cut off first number,
    q = new int[size-1];			// e.g., (2314) -> (213)
    for (i=1; i<size; i++) {
      if (p[i] < first)
        q[i-1] = p[i];
      else
        q[i-1] = p[i] - 1;
      }					
    sigma = new Permutation(size-1, q);
    
    if (sigma.next()) {				// advance 'tail', if possible
	  for (i=1; i<size; i++) {
	    if (sigma.p[i-1] < first)
	      p[i] = sigma.p[i-1];
	    else
	      p[i] = sigma.p[i-1] + 1;
	    }
	  }
	else
	  if (first+1 < size) {			// new first element, the
	    p[0] = ++first;				// rest are in increasing
	    for (i=0; i<size; i++) {	// order
	      if (i < first)
	        p[i+1] = i;
	      if (i > first)
	        p[i] = i;
	      }
	    }
	  else return false;
	return true;
	}
	
	
  public Permutation compose(Permutation s2) {		// multiplication of
    int i;								// permutations: the answer is 'this'
    Permutation sigma;					// left-multiplied by 's2' (i.e., 
    									// apply 's2' first, then 'this')
    if (size == s2.size)
      sigma = new Permutation(size); 	
    else return null;
    
    for (i=0; i<size; i++)
      sigma.p[i] = this.p[s2.p[i]];	
    
    return sigma;
    }
    
    
  public Permutation inverse() {		// return the inverse permutation
    int i;
    Permutation sigma = new Permutation(size);
    
    for (i=0; i<size; i++)
      sigma.p[this.p[i]] = i;
    return sigma;
    }
    
  
  public String toString() {			// printable form (gets called auto-
    int i;								//  matically by System.out.print)
    String str = new String("(");
    
    for (i=0; i<size-1; i++) {
      str = str.concat(String.valueOf(p[i] + 1));
      str = str.concat(",");
      }
    str = str.concat(String.valueOf(p[size-1] + 1));
    str = str.concat(")");
    return str;
    }
    
}
